--------------------------------------------------------------------------
--  DLX PROCESSOR MODEL SUITE
--  Copyright (C) 1995, Martin Gumm
--  University of Stuttgart / Department of Computer Science / IPVR-ISE
--------------------------------------------------------------------------
--  This program is free software; you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation; either version 1, or (at your option)
--  any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--------------------------------------------------------------------------
--  Last revision date : November 15 1995
--------------------------------------------------------------------------

--------------------------------------------------------------------------
--  *** SYNOPSYS synthesizable code (ver. 3.2.a) ***
--------------------------------------------------------------------------

--------------------------------------------------------------------------
--  Structural architecture of the controller
--  (decoder and instantiated state machine)
--
--  file controller2-dataflow.vhd
--------------------------------------------------------------------------

architecture dataflow of controller2 is

  component pipe_control
  port (phi1, phi2 : in std_logic;
        reset      : in std_logic;
        --
        stall_fetch  : out std_logic;
        stall_decode : out std_logic;
        stall_exec   : out std_logic;
        force_decode : out exc_type;
        --
        mem_memory_not_ready : in std_logic;
        fetch_memory_not_ready : in std_logic;
        --
        adr_ls2      : in std_logic_vector(0 to 1);
        rs1, rs2, rd : in dlx_reg_addr;
        load_instr   : in std_logic;
        reg_dest     : in std_logic;
        branch       : in std_logic;
        --
        mux1_out, mux2_out : out std_logic_vector(0 to 1);
        branch_reg_mux : out std_logic_vector(0 to 1));
  end component;

  component fetch
  port(phi1, phi2     : in std_logic;
       stall_fetch    : in std_logic;
       reset : in std_logic;
       ready           : in std_logic;
       memory_not_ready : out std_logic;
       fetch_mem_ctrl : out std_logic;
       ir_latch_en    : out std_logic;
       pc_latch       : out std_logic);
  end component;

 component decode
 port(stall, phi1   : in std_logic;
      instr_in      : in dlx_word;

-- shifted signals for execution stage

      smdr_latch_out : out std_logic;
      mdr_ctrl    : out std_logic_vector(0 to 2); -- lmdr_ctrl for mem stage
      s1_enab_out : out std_logic_vector(0 to 5);
      s2_enab_out : out std_logic_vector(0 to 3);
      alu_op_sel_out : out std_logic_vector(0 to 3);
      dest_enab_out : out std_logic_vector(0 to 4);
      const_sel_out : out std_logic_vector(0 to 1);
      immed_sel_out : out std_logic_vector(0 to 1);
      exc_enab_out : out std_logic_vector(0 to 7);
      icr_shifter_out : out std_logic;
      iar_mux_out : out std_logic;
      iar_latch_out : out std_logic;

-- shifted signals for memory stage

      mem_ctrl_out   : out std_logic_vector(0 to 5);
      entry_mux_out  : out std_logic;
      c_latch_out    : out std_logic;
      branch_out     : out std_logic;
      lmdr_latch_out : out std_logic;
      test_and_set_out : out test_and_set_instr;

-- shifted signals for write back stage

      wb_mux_out : out std_logic;
      rf_in_out : out std_logic;
      rd_out    : out dlx_reg_addr;

-- control signals of decode stage

      ir_1_latch : out std_logic;
      rs1_out : out dlx_reg_addr;
      rs2_out : out dlx_reg_addr;

-- return from datapath
      reg_value : in std_logic);

  end component;


  component exec
  port(stall, phi1, reset       : in std_logic;

-- shifted signals from decode stage

       mdr_ctrl_in : in std_logic_vector(0 to 2);
       s1_enab_in : in std_logic_vector(0 to 5);
       s2_enab_in : in std_logic_vector(0 to 3);
       alu_op_sel_in : in std_logic_vector(0 to 3);
       dest_enab_in : in std_logic_vector(0 to 4);
       const_sel_in : in std_logic_vector(0 to 1);
       immed_sel_in : in std_logic_vector(0 to 1);
       exc_enab_in : in std_logic_vector(0 to 7);
       icr_shifter_in : in std_logic;
       iar_mux_in : in std_logic;
       iar_latch_in : in std_logic;
       smdr_latch_in : in std_logic;
       --
       mem_ctrl_in : in std_logic_vector(0 to 5);
       entry_mux_in : in std_logic;
       c_latch_in : in std_logic;
       branch_in : in std_logic;
       lmdr_latch_in : std_logic;
       test_and_set_in : in test_and_set_instr;
       --
       wb_mux_in : in std_logic;
       rf_in_in : in std_logic;
       rd_in    : in dlx_reg_addr;

-- shifted signals for memory stage

       mem_ctrl_out   : out std_logic_vector(0 to 5);   -- memory control lines
       entry_mux_out  : out std_logic;
       c_latch_out    : out std_logic;

       pc_mux_out     : out std_logic;
       lmdr_latch_out : out std_logic;
       lmdr_ctrl_out  : out std_logic_vector(0 to 2);

-- shifted signals for write back stage

       wb_mux_out : out std_logic;
       rf_in_out  : out std_logic;
       rd_out     : out dlx_reg_addr;

-- control output lines

       s1_enab_out    : out std_logic_vector(0 to 5);    -- select s1 source
       s2_enab_out    : out std_logic_vector(0 to 3);    -- select s2_source
       alu_op_sel_out : out std_logic_vector(0 to 3);    -- alu operation
       dest_enab_out  : out std_logic_vector(0 to 4);    -- select destination
       const_sel_out  : out std_logic_vector(0 to 1);    -- select const for s1
       immed_sel_out  : out std_logic_vector(0 to 1);    -- select immed. from ir
       exc_enab_out   : out std_logic_vector(0 to 8);    -- enable set exc. bit
       icr_shifter_out : out std_logic;
       iar_mux_out : out std_logic;
       iar_latch_out : out std_logic;
       smdr_latch_out : out std_logic;
       smdr_ctrl_out : out std_logic_vector(0 to 2);
       test_and_set_mux : out std_logic_vector(0 to 1);
       --
       alu_neg : in std_logic;
       alu_zero : in std_logic);

  end component;


  component memory
  port(phi1, reset          : in std_logic;
       ready                : in std_logic;
       memory_not_ready : out std_logic;
       --
       mem_ctrl_in : in std_logic_vector(0 to 5);
       entry_mux_in : in std_logic;
       c_latch_in : in std_logic;

       pc_mux_in   : in std_logic;
       lmdr_latch_in : std_logic;
       lmdr_ctrl_in : in std_logic_vector(0 to 2);
       --
       --
       wb_mux_in : in std_logic;
       rf_in_in : in std_logic;
       rd_in    : in dlx_reg_addr;

-- shifted signals for write back stage

       wb_mux_out : out std_logic;
       rf_in_out  : out std_logic;
       rd_out     : out dlx_reg_addr;

-- control output lines

       adr_ls2        : in std_logic_vector(1 downto 0);
       mem_ctrl_out   : out std_logic_vector(0 to 7);   -- memory control lines
       entry_mux_out  : out std_logic;
       c_latch_out    : out std_logic;

       pc_mux_out     : out std_logic;
       lmdr_ctrl_out  : out std_logic_vector(0 to 2);
       lmdr_latch_out : out std_logic);

  end component;

  component wb
  port(phi1          : in std_logic;
       --
       rd_in         : in dlx_reg_addr;
       wb_mux_in     : in std_logic;
       rf_in_in      : in std_logic;

       rd_out	    : out dlx_reg_addr;
       wb_mux_sel    : out std_logic;
       rf_in_out     : out std_logic);

  end component;

signal intrn_ir_en : std_logic;
signal intrn_ir_1_en : std_logic;

signal s1_enab_ex : std_logic_vector(0 to 5);
signal s2_enab_ex : std_logic_vector(0 to 3);
signal alu_op_sel_ex : std_logic_vector(0 to 3);
signal dest_enab_ex : std_logic_vector(0 to 4);
signal const_sel_ex : std_logic_vector(0 to 1);
signal immed_sel_ex : std_logic_vector(0 to 1);
signal exc_enab_ex : std_logic_vector(0 to 7);
signal icr_shifter_ex : std_logic;
signal iar_mux_ex : std_logic;
signal iar_latch_ex : std_logic;
signal mem_ctrl_ex   : std_logic_vector(0 to 5);
signal entry_mux_ex  : std_logic;
signal c_latch_ex    : std_logic;
signal branch_ex   : std_logic;
signal mdr_sh_ctrl_ex : std_logic_vector(0 to 2);
signal lmdr_latch_ex : std_logic;
signal smdr_latch_ex : std_logic;
signal test_and_set_ex : test_and_set_instr;
signal wb_mux_ex : std_logic;
signal rf_in_ex  : std_logic;
signal rd_ex : dlx_reg_addr;

signal mem_ctrl_mem   : std_logic_vector(0 to 5);
signal entry_mux_mem  : std_logic;
signal c_latch_mem    : std_logic;
signal pc_mux_mem     : std_logic;
signal lmdr_latch_mem : std_logic;
signal lmdr_sh_ctrl_mem : std_logic_vector(0 to 2);
signal smdr_latch_mem : std_logic;
signal wb_mux_mem : std_logic;
signal rf_in_mem  : std_logic;
signal rd_mem : dlx_reg_addr;

signal wb_mux_wb : std_logic;
signal rf_in_wb  : std_logic;
signal rd_wb : dlx_reg_addr;

signal c_stall_fetch : std_logic;
signal c_stall_decode : std_logic;
signal c_stall_exec : std_logic;
--
signal intrn_alu_mux_1_sel : std_logic_vector(0 to 1);
signal intrn_alu_mux_2_sel : std_logic_vector(0 to 1);
--
signal intrn_a_b_latch_out : std_logic;
signal intrn_rs1_out : dlx_reg_addr;
signal intrn_rs2_out : dlx_reg_addr;
--
signal intrn_s1_enab : std_logic_vector(0 to 5);
signal intrn_s2_enab : std_logic_vector(0 to 3);
signal intrn_smdr_latch : std_logic;
signal intrn_dest_enab : std_logic_vector(0 to 4);
--
signal intrn_c_latch : std_logic;
signal intrn_lmdr_latch : std_logic;
--
signal intrn_rf_in_out : std_logic;
--
signal intrn_load_instr : std_logic;
signal intrn_reg_dest : std_logic;
--
signal intrn_fetch_memory_not_ready : std_logic;
signal intrn_mem_memory_not_ready : std_logic;
--
signal intrn_force_decode : exc_type;

begin

  -- instantiation of the pipeline control unit
  --
  control_unit : pipe_control
    port map (phi1 => phi1,
              phi2 => phi2,
              reset => reset,
              --
              stall_fetch => c_stall_fetch,
              stall_decode => c_stall_decode,
              stall_exec => c_stall_exec,
              force_decode => intrn_force_decode,
              --
              mem_memory_not_ready => intrn_mem_memory_not_ready,
              fetch_memory_not_ready => intrn_fetch_memory_not_ready, 
              --
              adr_ls2 => adr_ls2_in,
              rs1 => intrn_rs1_out,
              rs2 => intrn_rs2_out,
              rd => rd_ex,
              load_instr => intrn_load_instr,
              reg_dest => intrn_reg_dest,
              branch => branch_ex,
              --
              mux1_out => intrn_alu_mux_1_sel,
              mux2_out => intrn_alu_mux_2_sel,
              branch_reg_mux => reg_zero_sel); 

              
  -- instantiation of the instruction fetch stage
  --
  fetch_stage : fetch
    port map (phi1 => phi1, 
              phi2 => phi2,
              stall_fetch => c_stall_fetch,
              ready => ready_instr,
              reset => reset,
              memory_not_ready => intrn_fetch_memory_not_ready, 
              fetch_mem_ctrl => fetch_mem_ctrl, 
              ir_latch_en => intrn_ir_en,
              pc_latch => pc_latch);

  --
  -- decode stage
  -- produces all control signals
  --
  decode_stage : decode
    port map (stall => c_stall_decode, 
              phi1 => phi1,
              instr_in => instr_in, 
              --
              mdr_ctrl => mdr_sh_ctrl_ex,
              s1_enab_out => s1_enab_ex, 
              s2_enab_out => s2_enab_ex, 
              alu_op_sel_out => alu_op_sel_ex, 
              dest_enab_out => dest_enab_ex, 
              const_sel_out => const_sel_ex, immed_sel_out => immed_sel_ex,
              exc_enab_out => exc_enab_ex, icr_shifter_out => icr_shifter_ex,
              iar_mux_out => iar_mux_ex, iar_latch_out => iar_latch_ex,
              smdr_latch_out => smdr_latch_ex,
              mem_ctrl_out => mem_ctrl_ex, 
              entry_mux_out => entry_mux_ex, 
              c_latch_out => c_latch_ex, 
              branch_out => branch_ex, 
              lmdr_latch_out => lmdr_latch_ex, 
              test_and_set_out => test_and_set_ex, 
              wb_mux_out => wb_mux_ex,
              rf_in_out => rf_in_ex,
              rs1_out => intrn_rs1_out,
              rs2_out => intrn_rs2_out, 
              rd_out => rd_ex,
              ir_1_latch => intrn_ir_1_en,
              reg_value => reg_zero); 
  --
  -- execution stage
  -- 
  --
  exec_stage : exec
    port map (stall => c_stall_exec, phi1 => phi1, reset => reset, 
              mdr_ctrl_in => mdr_sh_ctrl_ex,
              s1_enab_in => s1_enab_ex, 
              s2_enab_in => s2_enab_ex,
              alu_op_sel_in => alu_op_sel_ex, dest_enab_in => dest_enab_ex,
              const_sel_in => const_sel_ex, immed_sel_in => immed_sel_ex,
              exc_enab_in => exc_enab_ex, icr_shifter_in => icr_shifter_ex,
              iar_mux_in => iar_mux_ex, iar_latch_in => iar_latch_ex,
              smdr_latch_in => smdr_latch_ex, mem_ctrl_in => mem_ctrl_ex,
              entry_mux_in => entry_mux_ex, c_latch_in => c_latch_ex,
              branch_in => branch_ex, 
              lmdr_latch_in => lmdr_latch_ex,
              test_and_set_in => test_and_set_ex, 
              wb_mux_in => wb_mux_ex,
              rf_in_in => rf_in_ex, 
              rd_in => rd_ex,
              -- 
              mem_ctrl_out => mem_ctrl_mem, 
              entry_mux_out => entry_mux_mem,
              c_latch_out => c_latch_mem, 
              pc_mux_out => pc_mux_mem,
              lmdr_latch_out => lmdr_latch_mem,
              lmdr_ctrl_out => lmdr_sh_ctrl_mem, 
              wb_mux_out => wb_mux_mem,
              rf_in_out => rf_in_mem, 
              rd_out => rd_mem, 
              s1_enab_out => intrn_s1_enab,
              s2_enab_out => intrn_s2_enab, 
              alu_op_sel_out => alu_op_sel,
              dest_enab_out => intrn_dest_enab, 
              const_sel_out => const_sel,
              immed_sel_out => immed_sel, exc_enab_out => exc_enab,
              icr_shifter_out => icr_shifter, iar_mux_out => iar_mux, 
              iar_latch_out => iar_latch, 
              smdr_latch_out => intrn_smdr_latch,
              smdr_ctrl_out => smdr_ctrl,
              test_and_set_mux(0) => t_and_s,
              test_and_set_mux(1) => test_result,
              alu_zero => alu_zero, 
              alu_neg => alu_neg); 
  --
  -- memory stage
  -- 
  --
  memory_stage : memory
    port map (phi1 => phi1,
              reset => reset, 
              ready => ready_data,
              memory_not_ready => intrn_mem_memory_not_ready,
              mem_ctrl_in => mem_ctrl_mem, 
              entry_mux_in => entry_mux_mem, 
              c_latch_in => c_latch_mem, 
              pc_mux_in => pc_mux_mem, 
              lmdr_latch_in => lmdr_latch_mem,
              lmdr_ctrl_in => lmdr_sh_ctrl_mem,
              wb_mux_in => wb_mux_mem, rf_in_in => rf_in_mem, rd_in => rd_mem,
              wb_mux_out => wb_mux_wb, rf_in_out => rf_in_wb, rd_out => rd_wb,
              mem_ctrl_out => mem_ctrl, entry_mux_out => entry_mux_sel,
              c_latch_out => intrn_c_latch,
              pc_mux_out => pc_mux_sel, 
              lmdr_latch_out => intrn_lmdr_latch,
              lmdr_ctrl_out => lmdr_ctrl,
              adr_ls2 => adr_ls2_in);
  --
  -- write back stage
  --
  wb_stage : wb
    port map (phi1 => phi1,
              rd_in => rd_wb, 
              wb_mux_in => wb_mux_wb, rf_in_in => rf_in_wb,
              wb_mux_sel => wb_mux_sel,
              rd_out => rd_out,
              rf_in_out => intrn_rf_in_out);

  --
  --
  alu_mux_1_sel(0) <= (intrn_alu_mux_1_sel(0) and intrn_s1_enab(4)) and 
                       intrn_s1_enab(5);
  alu_mux_2_sel(0) <= ((intrn_alu_mux_2_sel(0) and intrn_smdr_latch) 
                       and intrn_s2_enab(1)) and intrn_s2_enab(3);
  alu_mux_1_sel(1) <= (intrn_alu_mux_1_sel(1) and intrn_s1_enab(4)) and 
                       intrn_s1_enab(5);
  alu_mux_2_sel(1) <= ((intrn_alu_mux_2_sel(1) and intrn_smdr_latch) 
                       and intrn_s2_enab(1)) and intrn_s2_enab(3);

  smdr_mux_sel(0) <= intrn_alu_mux_2_sel(0);
  smdr_mux_sel(1) <= intrn_alu_mux_2_sel(1);
  --
  ir_latch_en_out <= intrn_ir_en; -- or not phi1;
  ir_1_latch <= intrn_ir_1_en; -- or not phi1;
  a_b_latch_out <= not phi2;
  rs1_out <= intrn_rs1_out;
  rs2_out <= intrn_rs2_out;
  --
  smdr_latch <= intrn_smdr_latch or not phi1;
  --
  dest_trigger : for i in 0 to 4 generate
   dest_enab(i) <= intrn_dest_enab(i) or not phi1;
  end generate;
  --
  lmdr_latch <= intrn_lmdr_latch or not phi1;

  c_latch <= intrn_c_latch or not phi1;
  --
  rf_in_out <= intrn_rf_in_out or not phi1;
  --
  s1_enab <= intrn_s1_enab;
  s2_enab <= intrn_s2_enab;
  --
  intrn_load_instr <= not lmdr_latch_ex;
  intrn_reg_dest <= not dest_enab_ex(0);
  --
end dataflow;


